import logging

from flask import Blueprint, request
from ckanext.vocabularies.logic import update_vocabulary_dsc

log = logging.getLogger("ckanext.vocabularies.dsc.subscribe")

vocabularies = Blueprint(
    'vocabularies',
    __name__
)

@vocabularies.route('/vocabularies/actions/update', methods=['GET'])
def update_vocabulary():
    url = request.args.get("url", type=str)
    type = request.args.get("type", type=str)
    if type is not None and url is not None:
        log.info("Updating vocabulary of type %s and url:%s",  type, url)
        if type == "dsc":
            update_vocabulary_dsc(url)
        elif type == "sparql":
            pass
        elif type == "turtle":
            pass
    return "Update complete."



#
# Copyright 2020 Fraunhofer Institute for Software and Systems Engineering
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import requests
import json

# Suppress ssl verification warning
requests.packages.urllib3.disable_warnings()


class ResourceApi:
    session = None
    recipient = None

    def __init__(self, recipient, auth=("admin", "password")):
        self.session = requests.Session()
        self.session.auth = auth
        self.session.verify = False

        self.recipient = recipient

    def create_catalog(self, data={}):
        response = self.session.post(self.recipient + "/api/catalogs", json=data)
        return response.headers["Location"]

    def create_offered_resource(self, data={}):
        response = self.session.post(self.recipient + "/api/offers", json=data)
        return response.headers["Location"]

    def create_representation(self, data={}):
        response = self.session.post(self.recipient + "/api/representations", json=data)
        return response.headers["Location"]

    def create_artifact(self, data={"value": "SOME LONG VALUE"}):
        response = self.session.post(self.recipient + "/api/artifacts", json=data)
        return response.headers["Location"]

    def update_artifact(self, artifact, data) -> bool:
        response = self.session.put(artifact, json=data)
        return response.status_code == 204

    def create_contract(
            self,
            data={
                "start": "2021-04-06T13:33:44.995+02:00",
                "end": "2021-12-06T13:33:44.995+02:00",
            },
    ):
        response = self.session.post(self.recipient + "/api/contracts", json=data)
        return response.headers["Location"]

    def create_rule(
            self,
            data={
                "value": """{
            "@context" : {
                "ids" : "https://w3id.org/idsa/core/",
                "idsc" : "https://w3id.org/idsa/code/"
            },
            "@type": "ids:Permission",
            "@id": "https://w3id.org/idsa/autogen/permission/cf1cb758-b96d-4486-b0a7-f3ac0e289588",
            "ids:action": [
                {
                "@id": "idsc:USE"
                }
            ],
            "ids:description": [
                {
                "@value": "provide-access",
                "@type": "http://www.w3.org/2001/XMLSchema#string"
                }
            ],
            "ids:title": [
                {
                "@value": "Example Usage Policy",
                "@type": "http://www.w3.org/2001/XMLSchema#string"
                }
            ]
            }"""
            },
    ):
        response = self.session.post(self.recipient + "/api/rules", json=data)
        return response.headers["Location"]

    def add_resource_to_catalog(self, catalog, resource):
        return self.session.post(
            catalog + "/offers", json=self.toListIfNeeded(resource)
        )

    def add_catalog_to_resource(self, resource, catalog):
        return self.session.post(
            resource + "/catalogs", json=self.toListIfNeeded(catalog)
        )

    def add_representation_to_resource(self, resource, representation):
        return self.session.post(
            resource + "/representations", json=self.toListIfNeeded(representation)
        )

    def add_artifact_to_representation(self, representation, artifact):
        return self.session.post(
            representation + "/artifacts", json=self.toListIfNeeded(artifact)
        )

    def add_contract_to_resource(self, resource, contract):
        return self.session.post(
            resource + "/contracts", json=self.toListIfNeeded(contract)
        )

    def add_rule_to_contract(self, contract, rule):
        return self.session.post(contract + "/rules", json=self.toListIfNeeded(rule))

    def toListIfNeeded(self, obj):
        if isinstance(obj, list):
            return obj
        else:
            return [obj]

    def get_data(self, artifact):
        return self.session.get(artifact + "/data")

    def get_artifacts_for_agreement(self, agreement):
        return json.loads(self.session.get(agreement + "/artifacts").text)


from SPARQLWrapper import SPARQLWrapper, JSON
import requests
import json
import logging

import ckan.plugins.toolkit as toolkit
from rdflib import plugin, URIRef
from rdflib.graph import Graph
#from rdflib.store import Store
#from rdflib_sqlalchemy import registerplugins

from ckanext.vocabularies.dsc.subscribe import Subscription

log = logging.getLogger("ckanext.vocabularies")



# deprecated
## create local triplestore on postgresql
#registerplugins()
#store = plugin.get("SQLAlchemy", Store)(identifier="vocabularies_store")

# keeps graphs in memory
graphs = dict()


def query_with_in_scheme(concept_scheme, lang="en"):
    if concept_scheme != None:
        in_scheme_triple = "?concept skos:inScheme <%(concept_scheme)s>"%dict(concept_scheme=concept_scheme)
    else: in_scheme_triple = ''
    return """
            PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
            select ?concept ?prefLabel (GROUP_CONCAT(?broaderPrefLabel;SEPARATOR="|") as ?path) (count(?broader) as ?counter) 
            where {
                ?concept a skos:Concept ;
                         skos:prefLabel ?prefLabel .
                %(in_scheme_triple)s         
                OPTIONAL { ?concept skos:broader+ ?broader .
                           ?broader skos:prefLabel ?broaderPrefLabel .
                        }
                FILTER ( langMatches( lang(?prefLabel), "%(lang)s") && langMatches( lang(?broaderPrefLabel), "%(lang)s"))       
                }
                group by ?concept ?prefLabel
            """%dict(in_scheme_triple=in_scheme_triple, lang=lang)


def query_poolparty(sparql_endpoint, query):
    response = requests.post(sparql_endpoint, data={"query": query, "format": "application/json"}, headers={ "Content-Type": "application/x-www-form-urlencoded"}).content.decode("utf-8")
    return json.loads(response)


def query_public_endpoint(sparql_endpoint, query):
    sparql = SPARQLWrapper(sparql_endpoint)
    sparql.setQuery(query)
    sparql.setReturnFormat(JSON)
    return sparql.queryAndConvert()


def format_results(results, sort_field: str = "label"):
    skos_choices = []
    if results['results']['bindings']:
        bindings = results['results']['bindings']
        for binding in bindings:
            path = (binding['path']['value'].split("|"))
            prefLabel = binding['prefLabel']['value']
            if path[0] != '':
                path.reverse()
                path.append(prefLabel)
                label = ' -> '.join(path)
            else:
                label = prefLabel
            choice = {
                'value': binding['concept']['value'],
                'label': label
            }
            skos_choices.append(choice)
        log.info('reformatted results')
        return sorted(skos_choices, key=lambda d: d[sort_field])
    else:
        return skos_choices.append({'value': '', 'label': 'empty'})


def query_dsc_resource(dsc_resource, dsc_resource_contract,query):
    if dsc_resource in graphs:
        graph = graphs[dsc_resource]
        log.info('Graph exists in triplestore...querying')
    else:
        log.info('Graph does not exist on triplestore...adding triples')
        graph = Graph()
        subscription = Subscription(dsc_resource, dsc_resource_contract)
        subscription.make_agreement()
        data = subscription.consume_resource()

        graph.parse(data=data, format="text/turtle")
        graphs[dsc_resource] = graph
    results = json.loads(graph.query(query).serialize(format='json'))
    return results

# deprecated
# def query_dsc_resource_rdbms(dsc_resource, query):
#     graph = Graph(store, identifier=dsc_resource)
#     try:
#         graph.open(URIRef(toolkit.config.get('ckanext.vocabularies.triplestore')), create=False)
#         log.info('Graph exists in triplestore...querying')
#     except RuntimeError as error:
#         log.info('Graph does not exist on triplestore...adding triples')
#         print(error)
#         graph = load_triples(dsc_resource, graph)
#     results = json.loads(graph.query(query).serialize(format='json'))
#     graph.close()
#     return results

# deprecated, keeping in case we need to use external triplestore
def load_triples(dsc_resource, graph):
    log.info('Opening store')
    graph.open(URIRef(toolkit.config.get('ckanext.vocabularies.triplestore')), create=True)
    log.info('Adding triples')
    graph.parse("test.ttl", format="text/turtle")
    log.info('Finished adding triples')
    return graph


def skos_choices_sparql_helper(field):
    '''Return a list of the concepts of a concept scheme served from a SPARQL endpoint'''

    sparql_endpoint = field.get('skos_choices_sparql_endpoint', None)
    is_poolparty = field.get('skos_choices_is_poolparty', False)
    concept_scheme = field.get('skos_choices_concept_scheme', None)
    dsc_resource = field.get('skos_choices_dsc_resource', None)
    dsc_resource_contract = field.get('skos_choices_dsc_resource_contract_offer', None)

    query = query_with_in_scheme(concept_scheme)
    if dsc_resource is not None:
        log.info('querying dsc')
        results = query_dsc_resource(dsc_resource, dsc_resource_contract, query)
    elif is_poolparty:
        results = query_poolparty(sparql_endpoint, query)
    else:
        results = query_public_endpoint(sparql_endpoint, query)

    log.info('received results')
    return format_results(results)


def skos_choices_get_label_by_value(formated_results: list, value: str):
    for choice in formated_results:
        if choice["value"] == value:
            return choice["label"]
    return value




import logging
from ckanext.ids.model import IdsResource, IdsAgreement
from ckanext.vocabularies.dsc.subscribe import Subscription
from ckanext.vocabularies.helpers import graphs
from rdflib import Graph

log = logging.getLogger("ckanext.vocabularies.logic")


def update_vocabulary_dsc(dsc_resource):
    local_resource = IdsResource.get(dsc_resource)
    local_agreement = local_resource.get_agreements()[0]
    log.info("Found the agreement url:" + local_agreement.id)
    subscription = Subscription(dsc_resource, None)
    subscription.agreement_url = local_agreement.id
    data = subscription.consume_resource()
    graph = Graph()
    graph.parse(data=data, format="text/turtle")
    log.info("Updating graph in local store...")
    graphs[dsc_resource] = graph
    log.info("Finished")



import ckan.plugins as plugins
from ckanext.vocabularies.helpers import skos_choices_sparql_helper
import ckanext.vocabularies.blueprints as blueprints
import ckan.plugins.toolkit as toolkit
import ckan.model as model
import ckan.logic as logic

class VocabulariesPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IConfigurer)
    # IConfigurer

    def find_or_create_dsc_user(self):
        username = "dsc_user_read"
        password = toolkit.config.get('ckanext.ids.trusts_local_dataspace_connector_password')
        email = "dsc_user_read@example.com"
        user = model.User.by_name(username)
        print("Finding dsc_user_read user")
        if not user:
            print("User not found. Creating one.")
            new_user = {
                "name": username,
                "email": email,
                "password": password
            }
            site_user = logic.get_action(u'get_site_user')({
                u'model': model,
                u'ignore_auth': True},
                {}
            )
            context = {
                u'model': model,
                u'session': model.Session,
                u'ignore_auth': True,
                u'user': site_user['name'],
            }
            toolkit.get_action("user_create")(context, new_user)
            user = model.User.by_name(username)
        else:
            print("User already exists...")

    def update_config(self, config_):
        toolkit.add_template_directory(config_, 'templates')
        toolkit.add_public_directory(config_, 'public')
        toolkit.add_resource('assets',
            'vocabularies')
        SQLALCHEMY_URL = toolkit.config.get('sqlalchemy.url').replace("postgresql", "postgresql+psycopg2")
        toolkit.config.store.update({'ckanext.vocabularies.triplestore': SQLALCHEMY_URL})
        self.find_or_create_dsc_user()

    # Declare that this plugin will implement ITemplateHelpers.
    plugins.implements(plugins.ITemplateHelpers)

    def get_helpers(self):
        '''Register the skos_vocabulary_helper() function as a template
        helper function.
        '''
        # Template helper function names should begin with the name of the
        # extension they belong to, to avoid clashing with functions from
        # other extensions.
        return {'skos_vocabulary_helper': skos_choices_sparql_helper}

    # IBlueprint
    plugins.implements(plugins.IBlueprint)

    def get_blueprint(self):
        return [blueprints.vocabularies]


import logging

import ckan.plugins.toolkit as toolkit

import pathlib
from ckanext.ids.model import IdsResource, IdsAgreement, IdsSubscription
from urllib.parse import urlparse
from ckanext.vocabularies.dsc.resourceapi import ResourceApi
from ckanext.vocabularies.dsc.subscriptionapi import SubscriptionApi
from ckanext.vocabularies.dsc.idsapi import IdsApi

log = logging.getLogger("ckanext.vocabularies.dsc.subscribe")

consumerUrl = toolkit.config.get('ckanext.ids.trusts_local_dataspace_connector_url') + ":" + toolkit.config.get('ckanext.ids.trusts_local_dataspace_connector_port')
username = toolkit.config.get('ckanext.ids.trusts_local_dataspace_connector_username')
password = toolkit.config.get('ckanext.ids.trusts_local_dataspace_connector_password')
consumer_alias = consumerUrl

consumer = IdsApi(consumerUrl, auth=(username, password))
local_node = toolkit.config.get("ckanext.ids.trusts_local_dataspace_connector_url")


class Subscription:
    offer_url = None
    contract_url = None
    provider_alias = None
    agreement_url = None
    first_artifact = None
    remote_artifact = None
    data_source = None
    endpoint = None
    route = None


    def __init__(self, offer_url, contract_url):
        self.offer_url = offer_url
        self.contract_url = contract_url

        parsed_url = urlparse(offer_url)
        self.provider_alias = parsed_url.scheme + "://" + parsed_url.netloc

    # IDS
    # Call description
    def make_agreement(self):
        log.info("Making agreement...")
        offer = consumer.descriptionRequest(self.provider_alias + "/api/ids/data", self.offer_url)
        log.debug(offer)
        artifact = offer["ids:representation"][0]["ids:instance"][0]['@id']
        self.remote_artifact = artifact
        log.debug(artifact)
        ## Check if an agreement already exists
        local_resource = IdsResource.get(self.offer_url)
        if local_resource is not None:
            local_agreement = local_resource.get_agreements()[0]
            self.agreement_url = local_agreement.id
            local_subscription = local_agreement.get_subscriptions()
            if len(local_subscription) == 0:
                self.subscribe()
            log.info("Agreement was made before... getting agreement url:%s", self.agreement_url)
            return
        # else Negotiate contract
        obj = offer["ids:contractOffer"][0]["ids:permission"][0]
        obj["ids:target"] = artifact
        log.debug(obj)
        response = consumer.contractRequest(self.provider_alias + "/api/ids/data", self.offer_url, artifact, False, obj)
        log.debug(response)
        self.agreement_url = response["_links"]["self"]["href"]
        log.info("Agreement made... agreement url:%s", self.agreement_url)
        local_resource = IdsResource(self.offer_url)
        local_resource.save()
        local_agreement = IdsAgreement(id=self.agreement_url,
                                   resource=local_resource,
                                   user="admin")
        local_agreement.save()
        self.subscribe()

    def consume_resource(self):
        log.info("Consuming resource...")
        consumerResources = ResourceApi(consumerUrl)
        artifacts = consumerResources.get_artifacts_for_agreement(self.agreement_url)
        log.debug(artifacts)
        first_artifact = artifacts["_embedded"]["artifacts"][0]["_links"]["self"]["href"]
        self.first_artifact = first_artifact
        log.debug(first_artifact)
        data = consumerResources.get_data(first_artifact).text
        log.info("Data acquired successfully!")
        #log.debug(data)
        return data
# Consumer

    def subscribe(self):
        self.create_data_source()
        self.create_endpoint()
        self.link_endpoint_data_source()
        self.create_route()
        self.link_endpoint_route()
        local_agreement = IdsAgreement.get(self.agreement_url)
        consumerResources = ResourceApi(consumerUrl)
        artifacts = consumerResources.get_artifacts_for_agreement(self.agreement_url)
        log.debug(artifacts)
        first_artifact = artifacts["_embedded"]["artifacts"][0]["_links"]["self"]["href"]
        self.first_artifact = first_artifact

        # subscribe to the requested artifact
        consumerSub = SubscriptionApi(consumerUrl)
        data = {
            "title": "CKAN on " + local_node + "vocabulary subscription",
            "description": "",
            "target": self.first_artifact.replace(local_node + ":8282", "http://locahlost:8080"),
            "location": self.route.replace(local_node + ":8282", "http://localhost:8080"),
            "subscriber":  local_node + ":5000",
            "pushData": "false",
        }

        response = consumerSub.create_subscription(data=data)
        local_subscription = IdsSubscription(id=response, agreement=local_agreement, user=username)
        local_subscription.save()
        log.debug(response)

        ## this is used to create ids subscription
        ## subscribe to the remote offer
        data = {
            "title": "DSC on " + local_node + "vocabulary subscription",
            "description": "",
            "target": self.remote_artifact,
            "location": consumerUrl + "/api/ids/data",
            "subscriber": consumerUrl,
            "pushData": "true",
        }
        response = consumerSub.subscription_message(
            data=data, params={"recipient": self.provider_alias + "/api/ids/data"}
        )

    def create_data_source(self):
        data = {
            "authentication": {
                "key": "dsc_user_read",
                "value": toolkit.config.get('ckanext.ids.trusts_local_dataspace_connector_password')
            },
            "type": "REST"
        }
        data_source = consumer.create_data_source(data)
        self.data_source = data_source["id"]
        return data_source

    def create_endpoint(self):
        data = {
            "location": "http://local-ckan:5000/vocabularies/actions/update?type=dsc&url=" + self.offer_url,
            "type": "GENERIC"
        }
        endpoint = consumer.create_endpoint(data)
        self.endpoint = pathlib.PurePath(endpoint["_links"]["self"]["href"]).name
        return endpoint

    def link_endpoint_data_source(self):
        response = consumer.link_endpoint_datasource(self.endpoint, self.data_source)
        return response

    def create_route(self):
        data = {
            "title": "Vocabulary Update Route",
            "deploy": "Camel"
        }
        response = consumer.create_route(data)
        self.route = response["_links"]["self"]["href"]
        return response

    def link_endpoint_route(self):
        response = consumer.link_route_endpoint(pathlib.PurePath(self.route).name, self.endpoint)
        return response